@@ -1,13 +1,18 @@
-use anyhow::{Error, Result};
+#[global_allocator]
+static GLOBAL: mimalloc::MiMalloc = mimalloc::MiMalloc;
+
+use anyhow::{bail, Error, Result};
 use env_logger::Env;
 use flate2::read::MultiGzDecoder;
 use flate2::write::DeflateDecoder;
 use gzp::deflate::{Bgzf, Gzip, Mgzip, RawDeflate};
 use gzp::par::compress::Compression;
 use gzp::par::decompress::ParDecompressBuilder;
+use gzp::{BgzfSyncReader, MgzipSyncReader};
 use gzp::{ZBuilder, ZWriter};
 use lazy_static::lazy_static;
 use log::info;
+use std::collections::HashSet;
 use std::fs::File;
 use std::io::{self, BufReader, BufWriter, Read, Write};
 use std::path::PathBuf;
@@ -27,6 +32,18 @@ use snap::read::FrameDecoder;
 
 const BUFFERSIZE: usize = 64 * 1024;
 
+macro_rules! string_set {
+    ( $( $x:expr ),* ) => {  // Match zero or more comma delimited items
+        {
+            let mut temp_set = HashSet::new();  // Create a mutable HashSet
+            $(
+                temp_set.insert(String::from($x)); // Insert each item matched into the HashSet
+            )*
+            temp_set // Return the populated HashSet
+        }
+    };
+}
+
 lazy_static! {
     /// Return the number of cpus as an &str
     pub static ref NUM_CPU: String = num_cpus::get().to_string();
@@ -76,8 +93,17 @@ fn get_input(path: Option<PathBuf>) -> Result<Box<dyn Read + Send + 'static>> {
     Ok(reader)
 }
 
-/// Get a buffered output writer from stdout or a file
-fn get_output(path: Option<PathBuf>) -> Result<Box<dyn Write + Send + 'static>> {
+/// Get a buffered output writer from stdout or a file.
+///
+/// If input is_some and in_place is true and output is None, figure out the inplace name
+#[allow(clippy::unnecessary_unwrap)]
+fn get_output(
+    path: Option<PathBuf>,
+    input_file: Option<PathBuf>,
+    in_place: bool,
+    is_decompress: bool,
+    format: Format,
+) -> Result<Box<dyn Write + Send + 'static>> {
     let writer: Box<dyn Write + Send + 'static> = match path {
         Some(path) => {
             if path.as_os_str() == "-" {
@@ -86,7 +112,44 @@ fn get_output(path: Option<PathBuf>) -> Result<Box<dyn Write + Send + 'static>>
                 Box::new(BufWriter::with_capacity(BUFFERSIZE, File::create(path)?))
             }
         }
-        None => Box::new(BufWriter::with_capacity(BUFFERSIZE, io::stdout())),
+        None => {
+            // Create a file
+            if in_place && input_file.is_some() {
+                let input_file = input_file.unwrap();
+                let (ext, allowed) = format.get_extension();
+                if is_decompress {
+                    if let Some(found_ext) = input_file.extension().map(|x| x.to_string_lossy()) {
+                        if allowed.contains(&found_ext.to_string()) {
+                            let input_file_str = input_file.to_string_lossy();
+                            let stripped = input_file_str
+                                .strip_suffix(&format!(".{}", found_ext))
+                                .unwrap();
+                            Box::new(BufWriter::with_capacity(
+                                BUFFERSIZE,
+                                File::create(stripped)?,
+                            ))
+                        } else {
+                            bail!(
+                                "Extension on {:?} does not match expected of {:?}",
+                                input_file,
+                                ext
+                            )
+                        }
+                    } else {
+                        bail!(
+                            "No extension on {:?}, does not match expected of {:?}",
+                            input_file,
+                            ext
+                        )
+                    }
+                } else {
+                    let out = format!("{}.{}", input_file.to_string_lossy(), ext);
+                    Box::new(BufWriter::with_capacity(BUFFERSIZE, File::create(out)?))
+                }
+            } else {
+                Box::new(BufWriter::with_capacity(BUFFERSIZE, io::stdout()))
+            }
+        }
     };
     Ok(writer)
 }
@@ -129,6 +192,7 @@ impl Format {
         writer: W,
         num_threads: usize,
         compression_level: u32,
+        pin_at: Option<usize>,
     ) -> Box<dyn ZWriter>
     where
         W: Write + Send + 'static,
@@ -137,33 +201,39 @@ impl Format {
             Format::Gzip => ZBuilder::<Gzip, _>::new()
                 .num_threads(num_threads)
                 .compression_level(Compression::new(compression_level))
+                .pin_threads(pin_at)
                 .from_writer(writer),
             Format::Bgzf => ZBuilder::<Bgzf, _>::new()
                 .num_threads(num_threads)
                 .compression_level(Compression::new(compression_level))
+                .pin_threads(pin_at)
                 .from_writer(writer),
             Format::Mgzip => ZBuilder::<Mgzip, _>::new()
                 .num_threads(num_threads)
                 .compression_level(Compression::new(compression_level))
+                .pin_threads(pin_at)
                 .from_writer(writer),
             #[cfg(feature = "any_zlib")]
             Format::Zlib => ZBuilder::<Zlib, _>::new()
                 .num_threads(num_threads)
                 .compression_level(Compression::new(compression_level))
+                .pin_threads(pin_at)
                 .from_writer(writer),
             Format::RawDeflate => ZBuilder::<RawDeflate, _>::new()
                 .num_threads(num_threads)
                 .compression_level(Compression::new(compression_level))
+                .pin_threads(pin_at)
                 .from_writer(writer),
             #[cfg(feature = "snappy")]
             Format::Snap => ZBuilder::<Snap, _>::new()
                 .num_threads(num_threads)
                 .compression_level(Compression::new(compression_level))
+                .pin_threads(pin_at)
                 .from_writer(writer),
         }
     }
 
-    fn get_highest_allowd_compression_leval(&self) -> u32 {
+    fn get_highest_allowed_compression_level(&self) -> u32 {
         match self {
             Format::Gzip => 9,
             #[cfg(feature = "libdeflate")]
@@ -182,6 +252,37 @@ impl Format {
             Format::Snap => u32::MAX,
         }
     }
+
+    fn get_lowest_allowed_compression_level(&self) -> u32 {
+        match self {
+            Format::Gzip => 0,
+            #[cfg(feature = "libdeflate")]
+            Format::Bgzf => 1,
+            #[cfg(not(feature = "libdeflate"))]
+            Format::Bgzf => 0,
+            #[cfg(feature = "libdeflate")]
+            Format::Mgzip => 1,
+            #[cfg(not(feature = "libdeflate"))]
+            Format::Mgzip => 0,
+            #[cfg(feature = "any_zlib")]
+            Format::Zlib => 0,
+            Format::RawDeflate => 0,
+            // compression level is ignored
+            #[cfg(feature = "snappy")]
+            Format::Snap => 0,
+        }
+    }
+
+    fn get_extension(&self) -> (&'static str, HashSet<String>) {
+        match self {
+            Format::Gzip => ("gz", string_set!["gz"]),
+            Format::Bgzf => ("gz", string_set!["gz", "bgz"]),
+            Format::Mgzip => ("gz", string_set!["gz", "mgz"]),
+            Format::Zlib => ("zz", string_set!["zz", "z", "gz"]),
+            Format::RawDeflate => ("gz", string_set!["gz"]),
+            Format::Snap => ("sz", string_set!["sz", "snappy"]),
+        }
+    }
 }
 
 /// Compress and decompress files.
@@ -192,6 +293,12 @@ struct Opts {
     #[structopt(short, long)]
     output: Option<PathBuf>,
 
+    /// Perform the compression / decompression in place.
+    ///
+    /// **NOTE** this will remove the input file at completion.
+    #[structopt(short = "I", long)]
+    in_place: bool,
+
     /// Input file to read from, empty or "-" to read from stdin
     #[structopt(name = "FILE", parse(from_os_str))]
     file: Option<PathBuf>,
@@ -212,20 +319,41 @@ struct Opts {
     /// Flag to switch to decompressing inputs. Note: this flag may change in future releases
     #[structopt(short, long)]
     decompress: bool,
+
+    /// Specify the physical core to pin threads at.
+    ///
+    /// This can provide a significant performance improvement, but has the downside of possibly conflicting
+    /// with other pinned cores. If you are running multiple instances of `crabz` at once you can manually
+    /// space out the pinned cores.
+    ///
+    /// # Example
+    /// - Instance 1 has `-p 4 -P 0` set indicating that it will use 4 cores pinned at 0, 1, 2, 3
+    /// - Instance 2 has `-p 4 -P 4` set indicating that it will use 4 cores pinned at 4, 5, 6, 7
+    #[structopt(short = "P", long)]
+    pin_at: Option<usize>,
 }
 
 fn main() -> Result<()> {
     let opts = setup();
-    if opts.compression_level > opts.format.get_highest_allowd_compression_leval() {
+    if opts.compression_level > opts.format.get_highest_allowed_compression_level()
+        || opts.compression_level < opts.format.get_lowest_allowed_compression_level()
+    {
         return Err(Error::msg("Invalid compression level"));
     }
 
     if opts.decompress {
         if let Err(err) = run_decompress(
-            get_input(opts.file)?,
-            get_output(opts.output)?,
+            get_input(opts.file.clone())?,
+            get_output(
+                opts.output,
+                opts.file.clone(),
+                opts.in_place,
+                opts.decompress,
+                opts.format,
+            )?,
             opts.format,
             opts.compression_threads,
+            opts.pin_at,
         ) {
             if is_broken_pipe(&err) {
                 exit(0)
@@ -233,17 +361,32 @@ fn main() -> Result<()> {
             return Err(err);
         }
     } else if let Err(err) = run_compress(
-        get_input(opts.file)?,
-        get_output(opts.output)?,
+        get_input(opts.file.clone())?,
+        get_output(
+            opts.output,
+            opts.file.clone(),
+            opts.in_place,
+            opts.decompress,
+            opts.format,
+        )?,
         opts.format,
         opts.compression_level,
         opts.compression_threads,
+        opts.pin_at,
     ) {
         if is_broken_pipe(&err) {
             exit(0)
         }
         return Err(err);
     }
+
+    // Remove input file
+    if opts.in_place {
+        if let Some(file) = opts.file {
+            std::fs::remove_file(file)?;
+        }
+    }
+
     Ok(())
 }
 
@@ -254,6 +397,7 @@ fn run_compress<R, W>(
     format: Format,
     compression_level: u32,
     num_threads: usize,
+    pin_at: Option<usize>,
 ) -> Result<()>
 where
     R: Read,
@@ -265,7 +409,7 @@ where
         num_threads,
         compression_level
     );
-    let mut writer = format.create_compressor(output, num_threads, compression_level);
+    let mut writer = format.create_compressor(output, num_threads, compression_level, pin_at);
     io::copy(&mut input, &mut writer)?;
     writer.finish()?;
     Ok(())
@@ -277,12 +421,12 @@ fn run_decompress<R, W>(
     mut output: W,
     format: Format,
     num_threads: usize,
+    pin_at: Option<usize>,
 ) -> Result<()>
 where
     R: Read + Send + 'static,
     W: Write + Send + 'static,
 {
-    // TODO: make passing - look for stdin / stdout and create similar to gzip behaviour otherwise
     info!(
         "Decompressing ({}) with {} threads available.",
         format.to_string(),
@@ -297,13 +441,14 @@ where
         }
         Format::Bgzf => {
             if num_threads == 0 {
-                let mut reader = MultiGzDecoder::new(input);
+                let mut reader = BgzfSyncReader::new(input);
                 io::copy(&mut reader, &mut output)?;
                 output.flush()?;
             } else {
                 let mut reader = ParDecompressBuilder::<Bgzf>::new()
                     .num_threads(num_threads)
                     .unwrap()
+                    .pin_threads(pin_at)
                     .from_reader(input);
                 io::copy(&mut reader, &mut output)?;
                 output.flush()?;
@@ -312,13 +457,14 @@ where
         }
         Format::Mgzip => {
             if num_threads == 0 {
-                let mut reader = MultiGzDecoder::new(input);
+                let mut reader = MgzipSyncReader::new(input);
                 io::copy(&mut reader, &mut output)?;
                 output.flush()?;
             } else {
                 let mut reader = ParDecompressBuilder::<Mgzip>::new()
                     .num_threads(num_threads)
                     .unwrap()
+                    .pin_threads(pin_at)
                     .from_reader(input);
                 io::copy(&mut reader, &mut output)?;
                 output.flush()?;
